import asyncio

import click
import maya
import aiofiles


from aioipfabric.filters import parse_filter

from ipfnetcfgbu.domain_remover import make_domain_remover
from ipfnetcfgbu.config_model import ConfigModel
from ipfnetcfgbu.ipf import IPFabricClient
from ipfnetcfgbu import logging
from .root import cli, opt_config_file, WithConfigCommand


async def exec_backup(
    config: ConfigModel, snapshot, filters, start_date, end_date, dry_run, force
):
    ipf_cfg = config.ipfabric

    log = logging.get_logger()

    filters = filters or ipf_cfg.filters

    log.info("Fetching inventory from IP Fabric")

    ipf = IPFabricClient()
    await ipf.login()

    if not snapshot:
        log.info(f"Using IP Fabric snapshot: {ipf.snapshots[0]['id']}")

    else:
        # API NOTE: "Untitled" snapshots are stored as name=None
        q_name = None if snapshot == "Untitled" else snapshot

        log.info(f"Looking for IP Fabric snapshot: {snapshot}")

        if not (
            sn_rec := next(
                (rec for rec in ipf.snapshots if rec["name"] == q_name), None
            )
        ):
            log.error(f"IP Fabric snapshot not found: {snapshot}")
            return
        ipf.active_snapshot = sn_rec["id"]

    if filters:
        log.info(f"Filters: {filters}")
        ipf_filters = parse_filter(filters)

    else:
        log.warning("No device filtering specified")
        ipf_filters = None

    # obtain the device inventory that is matching the User provided filters.
    # This list of hostnames is used to filter the configuration files that are
    # requested.

    devices = await ipf.fetch_devices(columns=["hostname"], filters=ipf_filters)

    if not len(devices):
        log.warning("No devices matching filter")
        return

    hostnames = {rec["hostname"] for rec in devices}

    log.info(f"Inventory contains {len(hostnames)} devices")

    def device_filter(hashrec):
        return hashrec["hostname"] in hostnames

    end_date = end_date or start_date.snap("1d")

    since_ts = int(start_date.epoch * 1_000)
    before_ts = int(end_date.epoch * 1_000)

    timespan_str = f"START={start_date}, END={end_date}"

    if force is True:
        log.info(f"Backup all configs: {timespan_str}")
    else:
        log.info(f"Backup configs that have changed: {timespan_str}")

    as_hostname = (
        make_domain_remover(domain_names=ipf_cfg.strip_hostname_domains)
        if ipf_cfg.strip_hostname_domains
        else lambda x: x
    )

    config_dir = config.defaults.configs_dir

    async def save_config(rec, config_text):
        hostname = as_hostname(rec["hostname"]).lower().replace("/", "-")
        log.info(f"SAVE CONFIG FOR: {hostname}")
        cfg_f = config_dir.joinpath(hostname + ".cfg")
        async with aiofiles.open(cfg_f, "w+") as ofile:
            await ofile.write(config_text)

    res = await ipf.fetch_device_configs(
        since_ts=since_ts,
        before_ts=before_ts,
        on_config=save_config,
        device_filter=device_filter,
        all_configs=force,
        dry_run=dry_run,
    )

    log.info(f"Total devices: {len(res)}")
    logging.stop()
    await ipf.logout()


def as_maya(ctx, param, value):  # noqa
    if not value:
        return None

    try:
        dt = maya.when(value)
        return dt.snap("@d") if value == "today" else dt

    except ValueError as exc:
        ctx.fail(f"{exc.args[0][:-1]}: {value}")


@cli.command(name="backup", cls=WithConfigCommand)
@opt_config_file
@click.option(
    "--start-date",
    help="Identifies the starting timestamp date/time",
    metavar="<DATE-TIME>",
    callback=as_maya,
    default="today",
)
@click.option(
    "--end-date",
    help="Identifies the ending timestamp date/time",
    metavar="<DATE-TIME>",
    callback=as_maya,
)
@click.option(
    "--force",
    help="Force backup of all configs, not just those that changed",
    is_flag=True,
)
@click.option(
    "--dry-run", help="Use to see device list that would be backed up", is_flag=True
)
@click.option("--filters", help="Override the `filters` option in the config file")
@click.option(
    "--snapshot",
    help="IPF snapshot name to use, default is most recent",
)
@click.pass_obj
def cli_backup(obj, **opts):
    """
    Backup network configurations.
    """
    opts["config"] = obj["config"]
    asyncio.run(exec_backup(**opts))



import click

from ipfnetcfgbu import config as _config
from ipfnetcfgbu.vcs import git

from .root import cli, opt_config_file, get_spec_nameorfirst

opt_vcs_name = click.option("--name", help="vcs name as defined in config file")


@cli.group(name="vcs")
def cli_vcs():
    """
    Version Control System subcommands.
    """
    pass  # pragma: no cover


class VCSCommand(click.Command):
    def invoke(self, ctx):
        cfg_fileopt = ctx.params["config"]
        try:
            app_cfgs = ctx.obj["app_cfg"] = _config.load(fileio=cfg_fileopt)
            if not (spec := get_spec_nameorfirst(app_cfgs.git, ctx.params["name"])):
                err_msg = (
                    "No configuration file provided, required for vcs support"
                    if not cfg_fileopt
                    else f"No vcs config section found in configuration file: {cfg_fileopt.name}"
                )
                raise RuntimeError(err_msg)

            ctx.obj["vcs_spec"] = spec
            super().invoke(ctx)

        except Exception as exc:
            ctx.fail(exc.args[0])


@cli_vcs.command(name="prepare", cls=VCSCommand)
@opt_config_file
@opt_vcs_name
@click.pass_context
def cli_vcs_prepare(ctx, **_cli_opts):
    """
    Prepare your system with the VCS repo.

    This command is used to setup your `configs_dir` as the VCS repository
    so that when you execute the backup process the resulting backup files
    can be stored in the VCS system.
    """

    git.vcs_prepare(
        spec=ctx.obj["vcs_spec"], repo_dir=ctx.obj["app_cfg"].defaults.configs_dir
    )


@cli_vcs.command(name="save", cls=VCSCommand)
@opt_config_file
@opt_vcs_name
@click.option("--tag-name", help="tag-release name")
@click.pass_context
def cli_vcs_save(ctx, **cli_opts):
    """
    Save changes into VCS repository.

    After you have run the config backup process you will need to push those
    changes into the VCS repository.  This command performs the necesssary
    steps to add changes to the repository and set a release tag.  The release
    tag by default is the timestamp in the form of
    "<year><month><day>_<hour><minute><second>"
    """
    git.vcs_save(
        ctx.obj["vcs_spec"],
        repo_dir=ctx.obj["app_cfg"].defaults.configs_dir,
        tag_name=cli_opts["tag_name"],
    )


@cli_vcs.command(name="status", cls=VCSCommand)
@opt_config_file
@opt_vcs_name
@click.pass_context
def cli_vcs_status(ctx, **_cli_opts):
    """
    Show VCS repository status.

    This command will show the status of the `configs_dir` contents so that you
    will know what will be changed before you run the `vcs save` command.
    """
    output = git.vcs_status(
        spec=ctx.obj["vcs_spec"], repo_dir=ctx.obj["app_cfg"].defaults.configs_dir
    )
    print(output)



# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from pathlib import Path

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

import toml

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from ipfnetcfgbu import logging
from .config_model import ConfigModel
from pydantic import ValidationError

__all__ = ["load"]


def validation_errors(filepath, errors):
    sp_4 = " " * 4
    as_human = ["Configuration errors", f"{sp_4}File:[{filepath}]"]

    for _err in errors:
        loc_str = ".".join(map(str, _err["loc"]))
        as_human.append(f"{sp_4}Section: [{loc_str}]: {_err['msg']}")

    return "\n".join(as_human)


def load(*, filepath=None, fileio=None) -> ConfigModel:
    app_cfg = dict()

    if filepath:
        app_cfg_file = Path(filepath)
        fileio = app_cfg_file.open()

    if fileio:
        app_cfg = toml.load(fileio)

    logging.start(app_cfg)

    app_defaults = app_cfg.get("defaults")
    if not app_defaults:
        app_cfg["defaults"] = dict(credentials={})

    try:
        cfg_obj = ConfigModel.parse_obj(app_cfg)

    except ValidationError as exc:
        filepath = fileio.name if fileio else ""
        raise RuntimeError(validation_errors(filepath=filepath, errors=exc.errors()))

    configs_dir: Path = cfg_obj.defaults.configs_dir
    if not configs_dir.is_dir():
        configs_dir.mkdir()

    return cfg_obj



import re
import os
from typing import Optional, List, Dict
from os.path import expandvars
from itertools import chain
from pathlib import Path

from pydantic import (
    BaseModel,
    SecretStr,
    BaseSettings,
    FilePath,
    DirectoryPath,
    validator,
    root_validator,
)


__all__ = [
    "ConfigModel",
    "Credentials",
    "GitSpec",
]

_var_re = re.compile(
    r"\${(?P<bname>[a-z0-9_]+)}" r"|" r"\$(?P<name>[^{][a-z_0-9]+)", flags=re.IGNORECASE
)


class NoExtraBaseModel(BaseModel):
    class Config:
        extra = "forbid"


class EnvExpand(str):
    """
    When a string value contains a reference to an environment variable, use
    this type to expand the contents of the variable using os.path.expandvars.

    For example like:
        password = "$MY_PASSWORD"
        foo_password = "${MY_PASSWORD}_foo"

    will be expanded, given MY_PASSWORD is set to 'boo!' in the environment:
        password -> "boo!"
        foo_password -> "boo!_foo"
    """

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if found_vars := list(filter(len, chain.from_iterable(_var_re.findall(v)))):
            for var in found_vars:
                if (var_val := os.getenv(var)) is None:
                    raise ValueError(f'Environment variable "{var}" missing.')

                if not len(var_val):
                    raise ValueError(f'Environment variable "{var}" empty.')

            return expandvars(v)

        return v


class EnvSecretStr(EnvExpand, SecretStr):
    @classmethod
    def validate(cls, v):
        return SecretStr.validate(EnvExpand.validate(v))


class Credentials(NoExtraBaseModel):
    username: Optional[EnvExpand]
    password: Optional[EnvSecretStr]
    token: Optional[EnvSecretStr]

    @root_validator()
    def ensure_credentials(cls, values):
        if not any(
            (all((values.get("username"), values.get("password"))), values["token"])
        ):
            raise ValueError("Missing required credentials")
        return values


class FilePathEnvExpand(FilePath):
    """ A FilePath field whose value can interpolated from env vars """

    @classmethod
    def __get_validators__(cls):
        yield from EnvExpand.__get_validators__()
        yield from FilePath.__get_validators__()


class DirPathEnvExpand(DirectoryPath):
    """ A DirectoryPath field whose value can interpolated from env vars """

    @classmethod
    def __get_validators__(cls):
        yield from EnvExpand.__get_validators__()
        yield from DirectoryPath.__get_validators__()


class Defaults(NoExtraBaseModel, BaseSettings):
    configs_dir: DirPathEnvExpand

    @validator("configs_dir")
    def _configs_dir(cls, value):  # noqa
        return Path(value).absolute()


class IPFabricModel(NoExtraBaseModel):
    server_url: EnvExpand
    credentials: Credentials
    filters: Optional[str]
    strip_hostname_domains: Optional[List[str]]


class GitSpec(NoExtraBaseModel):
    name: Optional[str]
    repo: str
    email: Optional[str]
    username: Optional[EnvExpand]
    password: Optional[EnvExpand]
    token: Optional[EnvSecretStr]
    deploy_key: Optional[FilePathEnvExpand]
    deploy_passphrase: Optional[EnvSecretStr]

    @validator("repo")
    def validate_repo(cls, repo):  # noqa
        expected = ("https:", "git@")
        if not repo.startswith(expected):
            raise ValueError(
                f"Bad repo URL [{repo}]: expected to start with {expected}."
            )
        return repo

    @root_validator
    def enure_proper_auth(cls, values):
        req = ("token", "deploy_key", "password")
        auth_vals = list(filter(None, (values.get(auth) for auth in req)))
        auth_c = len(auth_vals)
        if not auth_c:
            raise ValueError(
                f'Missing one of required auth method fields: {"|".join(req)}'
            )

        if auth_c > 1:
            raise ValueError(f'Only one of {"|".join(req)} allowed')

        if values.get("deploy_passphrase") and not values.get("deploy_key"):
            raise ValueError("deploy_key required when using deploy_passphrase")

        return values


class ConfigModel(NoExtraBaseModel):
    defaults: Defaults
    ipfabric: IPFabricModel
    git: Optional[List[GitSpec]]
    logging: Optional[Dict]



"""

References
----------

Logging in asyncio applications
   https://bit.ly/36WWgrf
"""

from typing import Set
import sys
import asyncio
from queue import SimpleQueue as Queue

from logging.config import dictConfig
from logging import getLogger
import logging
import logging.handlers


__all__ = ["start", "stop", "get_logger"]


_g_quelgr_listener: logging.handlers.QueueListener


class LocalQueueHandler(logging.handlers.QueueHandler):
    def emit(self, record: logging.LogRecord) -> None:
        # Removed the call to self.prepare(), handle task cancellation
        try:
            self.enqueue(record)

        except asyncio.CancelledError:
            raise

        except asyncio.QueueFull:
            self.handleError(record)


def setup_logging_queue(logger_names) -> None:
    """
    Move log handlers to a separate thread.

    Replace all configured handlers with a LocalQueueHandler, and start a
    logging.QueueListener holding the original handlers.
    """
    global _g_quelgr_listener
    queue = Queue()
    handlers: Set[logging.Handler] = set()
    que_handler = LocalQueueHandler(queue)

    for lname in logger_names:
        lgr = logging.getLogger(lname)
        lgr.addHandler(que_handler)
        for h in lgr.handlers[:]:
            if h is not que_handler:
                lgr.removeHandler(h)
                handlers.add(h)

    _g_quelgr_listener = logging.handlers.QueueListener(
        queue, *handlers, respect_handler_level=True
    )
    _g_quelgr_listener.start()


def start(app_cfg):
    log_cfg = app_cfg.get("logging") or {}
    log_cfg["version"] = 1

    dictConfig(log_cfg)
    setup_logging_queue(log_cfg.get("loggers") or [])


def stop():
    _g_quelgr_listener.stop()
    sys.stdout.flush()


def get_logger():
    return getLogger(__package__)



from importlib import metadata
from pathlib import Path

import click
from first import first

from ipfnetcfgbu import config
from ipfnetcfgbu import consts

VERSION = metadata.version("ipf-netcfgbu")


# -----------------------------------------------------------------------------
#
#                           CLI Custom Click Commands
#
# -----------------------------------------------------------------------------


class WithConfigCommand(click.Command):
    def invoke(self, ctx):
        try:
            ctx.obj["config"] = config.load(fileio=ctx.params["config"])
            super().invoke(ctx)

        except Exception as exc:
            ctx.fail(str(exc))


def get_spec_nameorfirst(spec_list, spec_name=None):
    if not spec_list:
        return None

    if not spec_name:
        return first(spec_list)

    return first(spec for spec in spec_list if getattr(spec, "name", "") == spec_name)


# -----------------------------------------------------------------------------
#
#                                CLI Options
#
# -----------------------------------------------------------------------------


def check_for_default(ctx, opt, value):
    if value:
        return value

    if (fileobj := Path(consts.DEFAULT_CONFIG_FILE)).exists():
        return fileobj.open()

    ctx.fail("Missing configuration file")


opt_config_file = click.option(
    "-C",
    "--config",
    type=click.File(),
    callback=check_for_default,
    envvar=consts.ENV_CONFIG_FILE,
)

# -----------------------------------------------------------------------------
# Inventory Options
# -----------------------------------------------------------------------------

opt_limits = click.option(
    "--limit",
    "-l",
    multiple=True,
    help="limit devices",
)

opt_excludes = click.option(
    "--exclude",
    "-e",
    multiple=True,
    help="exclude devices",
)


# opt_batch = click.option(
#     "--batch",
#     "-b",
#     type=click.IntRange(1, 500),
#     help="inevntory record processing batch size",
# )


@click.group()
@click.version_option(version=VERSION)
def cli():
    pass  # pragma: no cover



from aioipfabric import IPFabricClient as _IPFabricClient
from aioipfabric.mixins.configs import IPFConfigsMixin

__all__ = ["IPFabricClient"]


class IPFabricClient(_IPFabricClient, IPFConfigsMixin):
    pass



"""
This file contains the Version Control System (VCS) integration
using Git as the backend.   The following functions are exported
for use:

   * vcs_prepare:
      Used to prepare the repo directory for VCS use.

   * vcs_save:
      Used to save files in the repo directory into VCS and tag the collection
      with a release tag.

   * vcs_status:
      Used to show the current target status of file changes.

"""
# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import Optional
import os
from urllib.parse import urlsplit
from pathlib import Path
from datetime import datetime

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

import pexpect

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from ipfnetcfgbu.logging import get_logger
from ipfnetcfgbu.config_model import GitSpec

git_bin = "git"


def tag_name_timestamp() -> str:
    """
    Create the tag name using the current time with
    format <year><month#><day#>_<24hr><minute><sec>
    """
    return datetime.now().strftime("%Y%m%d_%H%M%S")  # pragma: no cover


# -----------------------------------------------------------------------------
#
#                             Git VCS Entrypoints
#
# -----------------------------------------------------------------------------


def vcs_save(gh_cfg: GitSpec, repo_dir: Path, tag_name: Optional[str] = None) -> bool:
    logr = get_logger()
    logr.info(f"VCS update git: {gh_cfg.repo}")

    ghr = git_runner(gh_cfg, repo_dir)
    if not tag_name:
        tag_name = tag_name_timestamp()

    output = ghr.run("status")
    if "nothing to commit" in output:
        logr.info("VCS no changes, skipping")
        return False

    logr.info(f"VCS saving changes, tag={tag_name}")

    commands = (
        ("add -A", False),
        (f"commit -m {tag_name}", False),
        ("push", True),
        (f"tag -a {tag_name} -m {tag_name}", False),
        ("push --tags", True),
    )

    for cmd, req_auth in commands:
        ghr.run(cmd, req_auth)

    return True


def vcs_prepare(spec: GitSpec, repo_dir: Path):
    logr = get_logger()
    logr.info(f"VCS prepare git: {spec.repo}")

    ghr = git_runner(spec, repo_dir)
    ghr.git_init()
    ghr.git_pull()


def vcs_status(spec: GitSpec, repo_dir: Path):
    logr = get_logger()
    logr.info(
        f"""
VCS REMOTE REPO: {spec.repo}
      LOCAL DIR: {str(repo_dir)}
"""
    )

    ghr = git_runner(spec, repo_dir)
    return ghr.run("status")


# -----------------------------------------------------------------------------
#
#                      Git Runners to perform commands
#
# -----------------------------------------------------------------------------


class GitRunner(object):
    """
    The GitRunner class is used to peform the specific `git` command
    operations requested for the VCS use cases.
    """

    def __init__(self, config: GitSpec, repo_dir):
        self.user = config.username or os.environ["USER"]
        self.config = config
        self.repo_dir = repo_dir
        self.git_file = repo_dir.joinpath(".git", "config")

        parsed = urlsplit(config.repo)
        if parsed.scheme == "https":
            self.repo_url = f"https://{self.user}@{parsed.netloc}{parsed.path}"
        else:
            self.repo_url = config.repo

    @property
    def repo_exists(self):
        return self.git_file.exists()

    @property
    def is_dir_empty(self):
        return not any(self.repo_dir.iterdir())

    def run_noauth(self, cmd: str):
        """
        Run the git command that does not require any user authentication
        """
        output, rc = pexpect.run(
            command=f"{git_bin} {cmd}",
            withexitstatus=True,
            cwd=self.repo_dir,
            encoding="utf-8",
        )

        if rc != 0:
            raise RuntimeError(f"git {cmd} failed: %s" % output)

        return output

    # run with auth is an alias to be created by subclass if needed
    run_auth = run_noauth

    def run(self, cmd: str, authreq=False):
        return [self.run_noauth, self.run_auth][authreq](cmd)  # noqa

    def git_init(self):
        output = self.run("remote -v") if self.repo_exists else ""
        if self.repo_url not in output:
            commands = (("init", False), (f"remote add origin {self.repo_url}", False))

            for cmd, req_auth in commands:
                self.run(cmd, req_auth)

        self.git_config()

    def git_pull(self):
        self.run("pull origin master", authreq=True)

    def git_config(self):
        config = self.config

        config_opts = (
            ("user.email", config.email or self.user),
            ("user.name", self.user),
            ("push.default", "matching"),
        )

        for cfg_opt, cfg_val in config_opts:
            self.run(f"config --local {cfg_opt} {cfg_val}")

    def git_clone(self):
        self.run(f"clone {self.repo_url} {str(self.repo_dir)}", authreq=True)
        self.git_config()


class GitAuthRunner(GitRunner):
    """
    Git Runner that is used for either User/Password or Token cases
    """

    PASSWORD_PROMPT = "Password for"

    def _get_secret(self):
        return self.config.token.get_secret_value()

    def run_auth(self, cmd):
        output, rc = pexpect.run(
            command=f"{git_bin} {cmd}",
            cwd=self.repo_dir,
            withexitstatus=True,
            encoding="utf-8",
            events={self.PASSWORD_PROMPT: self._get_secret() + "\n"},
        )

        if rc != 0:
            raise RuntimeError(output)

        return output


class GitTokenRunner(GitAuthRunner):
    # use the default password prompt value
    pass


class GitDeployKeyRunner(GitRunner):
    """
    Git Runner used with deployment keys without passphrase
    """

    def git_config(self):
        super().git_config()
        ssh_key = str(Path(self.config.deploy_key).absolute())
        self.run(
            f"config --local core.sshCommand 'ssh -i {ssh_key} -o StrictHostKeyChecking=no'"
        )


class GitSecuredDeployKeyRunner(GitDeployKeyRunner, GitAuthRunner):
    """
    Git Runner used when deployment key has passphrase configured
    """

    PASSWORD_PROMPT = "Enter passphrase for key"

    def _get_secret(self):
        return self.config.deploy_passphrase.get_secret_value()


def git_runner(gh_cfg: GitSpec, repo_dir: Path) -> GitRunner:
    """
    Used to select the Git Runner based on the configuration file
    settings.
    """
    if gh_cfg.token:
        return GitTokenRunner(gh_cfg, repo_dir)

    elif gh_cfg.deploy_key:
        if not gh_cfg.deploy_passphrase:
            return GitDeployKeyRunner(gh_cfg, repo_dir)
        else:
            return GitSecuredDeployKeyRunner(gh_cfg, repo_dir)

    # Note: this is unreachable code since the config-model validation should
    # have ensured the proper fields exist in the spec.

    raise RuntimeError("Git config missing authentication settings")  # pragma: no cover
